<?php

namespace FiloBlu\Flow\Model\Parser;

/**
 * Class AbstractParser
 * @package FiloBlu\Flow\Model\Parser
 */
abstract class AbstractParser
{
    /**
     * @var
     */
    public $map;

    /**
     * @var
     */
    protected $_channel;

    /**
     * @var \Monolog\Logger
     */
    protected $_logger;

    /**
     * @var \Magento\Framework\App\ObjectManager
     */
    protected $_objectManager;

    /**
     * @var
     */
    private $attribute_set_id;

    /**
     * @var
     */
    private $configurable_attributes;

    /**
     * @var
     */
    private $websites;

    /**
     * @var
     */
    private $stores;

    /**
     * AbstractParser constructor.
     * @param \FiloBlu\Flow\Helper\LoggerProvider $loggerProvider
     */
    public function __construct(
        \FiloBlu\Flow\Helper\LoggerProvider $loggerProvider
    ) {
        $this->_logger = $loggerProvider->getLogger();
        $this->_objectManager = \Magento\Framework\App\ObjectManager::getInstance();
    }

    /**
     * @param $resource
     * @return mixed
     */
    abstract public function process($resource);

    /**
     * @param $config
     * @return $this
     */
    public function setConfig($config)
    {
        return $this;
    }

    /**
     * @param $maplocation
     * @return bool
     */
    public function loadMap($mapLocation)
    {
        $map = $this->_objectManager->create(\FiloBlu\Flow\Model\Map::class);
        $map->load($mapLocation);

        if ($map->initialized) {
            $this->map = $map;
            return true;
        }

        return false;
    }

    /**
     * Parse
     *
     * @param array $data
     * @param bool $key_mode
     * @throws \Exception
     */
    public function parse($data, $key_mode = false)
    {
        $fields = [];
        $parsed_data = [];

        if ($this->check_mapping_count && !$key_mode && !$this->map->checkData($data)) {
            throw new \Exception('Data don\'t match mapping count ' . count($data));
        }

        $channel = $this->getChannel();

        if ($this->map) {
            $fields = $this->map->getFields();
        }

        $joins = ['map' => []];
        $insertDatas = [];

        // Default behaviour
        if ($this->check_mapping_count && count($fields)) {
            foreach ($fields as $i => $field) {

                $fieldMap = $this->map->get($field);

                $key = ($key_mode) ? (isset($fieldMap['from']) ? $fieldMap['from'] : $i) : $i;
                if (array_key_exists(trim($key), $data)) {
                    $fieldData = $data[trim($key)];
                    $result = $this->executeMappingValidatorsAndFilters($field, $data, $fieldMap, $fieldData);
                    if ($result) {
                        $parsed_data[$field] = $result;

                        //if there are joins
                        if (isset($fieldMap['joins']) && is_array($fieldMap['joins'])) {
                            $joins['key'] = $field;
                            if(isset($fieldMap['joins'][0][$result])){
                                $joins['map'] = $fieldMap['joins'][0][$result];
                            }
                            //TODO: Log unmapped
                        }
                    }
                }
            }

            foreach ($joins['map'] as $item) {
                $p = $parsed_data;
                $p[$joins['key']] = $item;
                $insertDatas[] = $p;
            }

            if(empty($joins['map'])){
                $insertDatas[] = $parsed_data;
            }

            //Expand configurables if is set expand_configurables:1 in map
            if($this->map->expandConfigurables()){
                $toExpand = $insertDatas;
                $insertDatas = [];
                foreach($toExpand as $itemToExpand){
                    $helperProduct = $this->_objectManager->create('FiloBlu\Flow\Helper\Product');
                    $childs = $helperProduct->getConfigurableChildsBySku($itemToExpand['sku']);
                    
                    foreach($childs as $child) {
                        $itemToExpand['sku'] = $child;
                        $insertDatas[] = $itemToExpand;
                    }
                }                
            }

            // This is no check_mapping_count (Talend parse) then CSV headers will match Magento attributes
        } elseif (!$this->check_mapping_count) {

            // Merging headers with current data from the csv
            // array_combine gives a warning if arrays have not equal columns number (using foreach)
            foreach ($this->header as $header_key => $header_column_name) {
                $parsed_data[$header_column_name] = $data[$header_key];
                $fieldMap = [];

                if (count($fields)) {
                    $fieldMap = $this->map->get($header_column_name);
                }

                $field = $header_column_name;
                $fieldData = $data[$header_key];
                $result = $this->executeMappingValidatorsAndFilters($field, $data, $fieldMap, $fieldData);

                if ($result) {
                    $parsed_data[$field] = $result;
                }
            }
            $insertDatas[] = $parsed_data;
        }

        if (method_exists($channel, 'insertData') && !is_null($channel)) {
            foreach ($insertDatas as $toInsert){
                $channel->insertData($toInsert);
            }
        } else {
            throw new \Exception('No method insertData for class ' . get_class($channel));
        }
    }

    /**
     * Validate single field applying Validators
     * @return mixed
     */
    public function getChannel()
    {
        return $this->_channel;
    }

    /**
     *
     * @param $channel
     */
    public function setChannel($channel)
    {
        $this->_channel = $channel;
    }

    /**
     * @param $field
     * @param $data
     * @param $fieldMap
     * @param $fieldData
     * @return null
     * @throws \Exception
     */
    private function executeMappingValidatorsAndFilters($field, $data, $fieldMap, $fieldData)
    {
        if (isset($fieldMap['skip']) || !$fieldMap) {
            return null;
        }

        $filters = null;
        $validators = null;
        $value = null;

        if (isset($fieldMap['validators'])) {
            $validators = $fieldMap['validators'];
        }

        if (isset($fieldMap['filters'])) {
            $filters = $fieldMap['filters'];
        }

        if ($this->_validate($fieldData, $validators, $field)) {
            $value = $this->_filter($fieldData, $filters, $data);
            return $value;
        } else {
            throw new \Exception("{$value} is not Valid for {$field}");
        }

        return null;
    }

    /**
     * Parse Data
     *
     * @param $value
     * @param $validators
     * @param $field
     * @return bool
     */
    protected function _validate($value, $validators, $field)
    {
        if (!$validators || count($validators) == 0) {
            return true;
        }

        foreach ($validators as $validator) {
            $validatorClass = "FiloBlu\Bouncer\Model\Parser\Validators\\" . ucfirst($validator);
            $validator = new $validatorClass();

            if (!$validator->isValid($value)) {
                return false;
            }
        }

        return true;
    }

    /**
     * @param $value
     * @param $filters
     * @param $data
     * @return mixed
     */
    protected function _filter($value, $filters, $data)
    {
        if (!$filters || count($filters) == 0) {
            return $value;
        }

        foreach ($filters as $filterData) {
            $filterClass = "FiloBlu\Flow\Model\Filters\\" . ucfirst($filterData['class']);
            $filter = $this->_objectManager->create($filterClass);

            if (method_exists($filter, 'setRow')) {
                $filter->setRow($data);
            }

            $value = $filter->filter($value, $filterData['args']);
        }

        return $value;
    }
}
